"""
Import scraped Instagram data

It's prohibitively difficult to scrape data from Instagram within 4CAT itself
due to its aggressive rate limiting. Instead, import data collected elsewhere.
"""
import datetime
import re

from backend.lib.search import Search


class SearchInstagram(Search):
    """
    Import scraped Instagram data
    """
    type = "instagram-search"  # job ID
    category = "Search"  # category
    title = "Import scraped Instagram data"  # title displayed in UI
    description = "Import Instagram data collected with an external tool such as Zeeschuimer."  # description displayed in UI
    extension = "ndjson"  # extension of result file, used internally and in UI
    is_from_extension = True

    # not available as a processor for existing datasets
    accepts = [None]
    references = [
        "[Zeeschuimer browser extension](https://github.com/digitalmethodsinitiative/zeeschuimer)",
        "[Worksheet: Capturing TikTok data with Zeeschuimer and 4CAT](https://tinyurl.com/nmrw-zeeschuimer-tiktok) (also covers usage with Instagram)"
    ]

    # some magic numbers instagram uses
    MEDIA_TYPE_PHOTO = 1
    MEDIA_TYPE_VIDEO = 2
    MEDIA_TYPE_CAROUSEL = 8

    def get_items(self, query):
        """
        Run custom search

        Not available for Instagram
        """
        raise NotImplementedError("Instagram datasets can only be created by importing data from elsewhere")


    @staticmethod
    def map_item(item):
        """
        Map Instagram item

        Instagram importing is a little bit roundabout since we can expect
        input in two separate and not completely overlapping formats - an "edge
        list" or an "item list", and posts are structured differently between
        those, and do not contain the same data. So we find a middle ground
        here... each format has its own handler function

        :param dict item:  Item to map
        :return:  Mapped item
        """
        is_graph_response = "__typename" in item

        if is_graph_response:
            return SearchInstagram.parse_graph_item(item)
        else:
            return SearchInstagram.parse_itemlist_item(item)

    @staticmethod
    def parse_graph_item(node):
        """
        Parse Instagram post in Graph format

        :param node:  Data as received from Instagram
        :return dict:  Mapped item
        """
        try:
            caption = node["edge_media_to_caption"]["edges"][0]["node"]["text"]
        except IndexError:
            caption = ""

        num_media = 1 if node["__typename"] != "GraphSidecar" else len(node["edge_sidecar_to_children"]["edges"])

        # get media url
        # for carousels, get the first media item, for videos, get the video
        # url, for photos, get the highest resolution
        if node["__typename"] == "GraphSidecar":
            media_node = node["edge_sidecar_to_children"]["edges"][0]["node"]
        else:
            media_node = node

        if media_node["__typename"] == "GraphVideo":
            media_url = media_node["video_url"]
        elif media_node["__typename"] == "GraphImage":
            resources = media_node.get("display_resources", media_node.get("thumbnail_resources"))
            try:
                media_url = resources.pop()["src"]
            except AttributeError:
                media_url = media_node.get("display_url", "")
        else:
            media_url = media_node["display_url"]

        # type, 'mixed' means carousel with video and photo
        type_map = {"GraphSidecar": "photo", "GraphVideo": "video"}
        if node["__typename"] != "GraphSidecar":
            media_type = type_map.get(node["__typename"], "unknown")
        else:
            media_types = set([s["node"]["__typename"] for s in node["edge_sidecar_to_children"]["edges"]])
            media_type = "mixed" if len(media_types) > 1 else type_map.get(media_types.pop(), "unknown")

        location = {"name": "", "latlong": "", "city": ""}
        # location has 'id', 'has_public_page', 'name', and 'slug' keys in tested examples; no lat long or "city" though name seems
        if node.get("location"):
            location["name"] = node["location"].get("name")
            # Leaving this though it does not appear to be used in this type; maybe we'll be surprised in the future...
            location["latlong"] = str(node["location"]["lat"]) + "," + str(node["location"]["lng"]) if node[
                "location"].get("lat") else ""
            location["city"] = node["location"].get("city")

        mapped_item = {
            "id": node["shortcode"],
            "thread_id": node["shortcode"],
            "parent_id": node["shortcode"],
            "body": caption,
            "author": node["owner"]["username"],
            "timestamp": datetime.datetime.fromtimestamp(node["taken_at_timestamp"]).strftime("%Y-%m-%d %H:%M:%S"),
            "author_fullname": node["owner"].get("full_name", ""),
            "author_avatar_url": node["owner"].get("profile_pic_url", ""),
            "type": media_type,
            "url": "https://www.instagram.com/p/" + node["shortcode"],
            "image_url": node["display_url"],
            "media_url": media_url,
            "hashtags": ",".join(re.findall(r"#([^\s!@#$%ˆ&*()_+{}:\"|<>?\[\];'\,./`~']+)", caption)),
            # "usertags": ",".join(
            #     [u["node"]["user"]["username"] for u in node["edge_media_to_tagged_user"]["edges"]]),
            "num_likes": node["edge_media_preview_like"]["count"],
            "num_comments": node.get("edge_media_preview_comment", {}).get("count", 0),
            "num_media": num_media,
            "location_name": location["name"],
            "location_latlong": location["latlong"],
            "location_city": location["city"],
            "unix_timestamp": node["taken_at_timestamp"]
        }

        return mapped_item

    @staticmethod
    def parse_itemlist_item(node):
        """
        Parse Instagram post in 'item list' format

        :param node:  Data as received from Instagram
        :return dict:  Mapped item
        """
        num_media = 1 if node["media_type"] != SearchInstagram.MEDIA_TYPE_CAROUSEL else len(node["carousel_media"])
        caption = "" if not node.get("caption") else node["caption"]["text"]

        # get media url
        # for carousels, get the first media item, for videos, get the video
        # url, for photos, get the highest resolution
        if node["media_type"] == SearchInstagram.MEDIA_TYPE_CAROUSEL:
            media_node = node["carousel_media"][0]
        else:
            media_node = node

        if media_node["media_type"] == SearchInstagram.MEDIA_TYPE_VIDEO:
            media_url = media_node["video_versions"][0]["url"]
            display_url = media_node["image_versions2"]["candidates"][0]["url"]
        elif media_node["media_type"] == SearchInstagram.MEDIA_TYPE_PHOTO:
            media_url = media_node["image_versions2"]["candidates"][0]["url"]
            display_url = media_url
        else:
            media_url = ""
            display_url = ""

        # type, 'mixed' means carousel with video and photo
        type_map = {SearchInstagram.MEDIA_TYPE_PHOTO: "photo", SearchInstagram.MEDIA_TYPE_VIDEO: "video"}
        if node["media_type"] != SearchInstagram.MEDIA_TYPE_CAROUSEL:
            media_type = type_map.get(node["media_type"], "unknown")
        else:
            media_types = set([s["media_type"] for s in node["carousel_media"]])
            media_type = "mixed" if len(media_types) > 1 else type_map.get(media_types.pop(), "unknown")

        if "comment_count" in node:
            num_comments = node["comment_count"]
        elif "comments" in node and type(node["comments"]) is list:
            num_comments = len(node["comments"])
        else:
            num_comments = -1

        location = {"name": "", "latlong": "", "city": ""}
        if node.get("location"):
            location["name"] = node["location"].get("name")
            location["latlong"] = str(node["location"]["lat"]) + "," + str(node["location"]["lng"]) if node[
                "location"].get("lat") else ""
            location["city"] = node["location"].get("city")

        mapped_item = {
            "id": node["code"],
            "thread_id": node["code"],
            "parent_id": node["code"],
            "body": caption,
            "author": node["user"]["username"],
            "author_fullname": node["user"]["full_name"],
            "author_avatar_url": node["user"]["profile_pic_url"],
            "timestamp": datetime.datetime.fromtimestamp(node["taken_at"]).strftime("%Y-%m-%d %H:%M:%S"),
            "type": media_type,
            "url": "https://www.instagram.com/p/" + node["code"],
            "image_url": display_url,
            "media_url": media_url,
            "hashtags": ",".join(re.findall(r"#([^\s!@#$%ˆ&*()_+{}:\"|<>?\[\];'\,./`~']+)", caption)),
            # "usertags": ",".join(
            #     [u["node"]["user"]["username"] for u in node["edge_media_to_tagged_user"]["edges"]]),
            "num_likes": node["like_count"],
            "num_comments": num_comments,
            "num_media": num_media,
            "location_name": location["name"],
            "location_latlong": location["latlong"],
            "location_city": location["city"],
            "unix_timestamp": node["taken_at"]
        }

        return mapped_item
